#define STANDALONE
#define VS1053
#include <hardware.h>
#include <imalib.h>
#include <encoders.h>
#include <asmfuncs.h>
#include <stream.h>
#include <parametric.h>
#include <stdlib.h>
#include <string.h>

/*
  There are two modes:
  1.Stream anything up to 44.1kHz
  - If you stream mp3, the decoder will automatically synchronize to the
    data, and determine the number of channels and samplerate from the mp3
    frame headers.
  - If you stream something else than mp3, the decoder needs to receive the
    headers. This means the receiver (this software) needs to be started
    first, and the encoder (UartSender) afterwards.

  2.Stream mono IMA ADPCM up to 44.1kHz (enable the define STREAM_ADPCM below)
  - The decoder will expect mono IMA ADPCM with block align 256 (505 samples
    per frame), synchronizing to the IMA ADPCM headers, and determining
    the samplerate from the time between two IMA ADPCM blocks.
  - If you buffer more than one IMA ADPCM frame to be sent as a burst,
    you should disable samplerate detection.

  The UART baudrate is 460800bps. You can change it from vs1053b-uartin.s .
 */


//#define STREAM_ADPCM //stream mono IMA ADPCM at 44.1kHz

#define ENCODE_CHANNELS 1 //only 1 channel supported
#define ENCODE_RATE 44100

void putch(register __a0 int ch);
void putword(register __a0 int ch);

s_int32 fillFlt = 0;
extern __mem_y u_int32 curFctl;

void AdaptRate() {
    s_int16 diff = StreamDiff();
    s_int16 audiff = AudioBufFill()/2;
    //USEX(SCI_AICTRL0) = diff;
    //USEX(SCI_AICTRL1) = audiff;
#define GOOD_FILL (AUDIO_BUFFER_SZ/4) //AUDIO_BUFFER_SZ/3
    fillFlt = fillFlt - (fillFlt >> 7) + (audiff - GOOD_FILL/2) + diff;
    {
      u_int32 fillTmp = /*audioPtr.*/ curFctl + fillFlt/8;
        static s_int32 oldFlt = 0;
        if (labs(oldFlt - fillTmp) > 2) {
            oldFlt = fillTmp;
            USEX(FREQCTLL) = fillTmp;
            USEX(FREQCTLH) = (USEX(FREQCTLH) & 0xfff0) | ((u_int16)((u_int32)fillTmp>>16) & 15);

            //USEX(SCI_AICTRL2) = USEX(FREQCTLL);
            //USEX(SCI_AICTRL3) = USEX(FREQCTLH);
        }
    }
}
u_int32 MyCheckTag(register __reg_d u_int32 head) {
#ifdef STREAM_ADPCM
  /* If IMA ADPCM, then this is as good place as any to keep reading
     data and decoding. */
  static __mem_y s_int16 imaTmpY[4];
  static s_int16 imaTmpX[8], tmpBuf[16];
  s_int16 adpcmPhase = 0;
  s_int16 val = 0;
  s_int16 st = 0;
  USEX(VS1053_IROM4) = IROM4_ENABLE; /*Make sure some routines are visible*/
  SetHardware(ENCODE_CHANNELS, ENCODE_RATE); //the default
  /* IMA ADPCM decoding, (re)implement it here. */
  while (1) {
      if (adpcmPhase == 0) { 
        SearchForImaHeader(1); //Mono -- TODO: determine samplerate
        MyGetCPairs((__mem_y void *)imaTmpY, 4/2);
        AdaptRate(); /* Adapt at the header. */
        ImaExpandStart(&st, &val, (__mem_y void *)imaTmpY, (void *)imaTmpX);//from libimaadpcm.a
        tmpBuf[0] = imaTmpX[0];
        tmpBuf[1] = imaTmpX[0];
        WmaStereoCopy(tmpBuf, 1);
        adpcmPhase += 1;
      } else {
        MyGetCPairs((__mem_y void *)imaTmpY, 4/2);
        ImaExpandContinue(8, &st, &val, (__mem_y void *)imaTmpY, (void *)imaTmpX);//from libimaadpcm.a
        {
          register s_int16 *d = tmpBuf;
          register s_int16 *s = imaTmpX;
          register int i;
          for (i=0;i<8;i++) {
            *d++ = *s;
            *d++ = *s++;
          }
        }
        WmaStereoCopy(tmpBuf, 8);
        adpcmPhase += 8;
        if (adpcmPhase >= 505) { //fixed block-align 256, i.e. 505 samples.
          adpcmPhase = 0;
        }
      }
  }
  //#pragma msg 213 off
#pragma msg 36 off
#pragma msg 38 off
#else
  AdaptRate(); //if mp3, adapt at each mp3 header.
  return head;
#endif
}

#pragma msg 36 on
#pragma msg 38 on


u_int16 triggerCount = 0;
s_int16 ApplAddr(s_int16 register __i0 **d, s_int16 register __a1 mode,
        	 s_int16 register __a0 n) {
    USEX(INT_ENABLE) = 0x23; //enable DAC, SCI, RX
    /*If not mp3 but WAV, use some adaptation anyway.
      Less accurate because it doesn't happen on encoding block boundaries.*/
    if (mode == APPL_AUDIO && USEX(SCI_HDAT1) == 0x7665 /*'ve'*/) {
      triggerCount += n;
      if (triggerCount >= hwSampleRate / 32) {
        triggerCount = 0;
        MyCheckTag(0);
      }
    }
    return n;
}
