#define STANDALONE
#define VS1103
#include <hardware.h>
#include <imalib.h>
#include <stream.h>
#include <stdlib.h>
#include <string.h>

/*
  Stream mono IMA ADPCM up to 44.1kHz:
  - The decoder will expect mono IMA ADPCM with block align 256 (505 samples
    per frame), synchronizing to the IMA ADPCM headers, and determining
    the samplerate from the time between two IMA ADPCM blocks.
  - Performs samplerate finetuning to adapt to the transmitter's rate variation.
  - If you buffer more than one IMA ADPCM frame to be sent as a burst,
    you should disable samplerate detection.

  The UART baudrate is 460800bps. You can change it by calculating new
  divider in the vs1103() function.
 */

#define ENCODE_CHANNELS 1 //must be 1
#define ENCODE_RATE 8000 //44100 //default rate, automatically determines

void putch(register __a0 int ch);
void putword(register __a0 int ch);

s_int32 fillFlt = 0;
extern __mem_y u_int32 curFctl; //In VS1103 curFctl is in X!
#define PERIP32(x) (*(volatile u_int32 __mem_x *)(u_int32)(x))

void AdjustFilter(void) {
  s_int16 diff = StreamDiff();
  s_int16 audiff = AudioBufFill()/2;
#define GOOD_FILL (2048/*AUDIO_BUFFER_SZ*//2) //AUDIO_BUFFER_SZ/3
  fillFlt = fillFlt - (fillFlt >> 7) + (audiff - GOOD_FILL/2) + diff;
  {
    u_int32 fillTmp = /*audioPtr.*/ PERIP32(&curFctl) + fillFlt/16;
    static s_int32 oldFlt = 0;
    if (labs(oldFlt - fillTmp) > 2) {
      oldFlt = fillTmp;
      USEX(FREQCTLL) = fillTmp;
      USEX(FREQCTLH) = (USEX(FREQCTLH) & 0xfff0) | ((u_int16)((u_int32)fillTmp>>16) & 15);
    }
  }
}

/* Some things we need from vs1103b ROM. */
extern u_int16 skipMixer;
extern u_int16 __mem_y adpcmSource;
extern u_int16 adpcmPhase;
extern u_int16 adpcmTmpX[18];
extern u_int16 __mem_y adpcmTmpY[18];
auto s_int16 SdiGetC(void);
auto s_int16 AdpcmBufFill(void);
auto s_int16 AdpcmReadBytes(u_int16 *d, s_int16 byteOffset, int n);

void vs1103(void) {
  USEX(SCI_CLOCKF) = 0xc430; //4.0x 12.288MHz
  InitHardware();
  uartByteSpeed = 460800/10;
  SetClockFromSpi();
  InitCommon();
  applAddr = NULL;
  stream_wr_pointer = stream_rd_pointer = (void *)stream_buffer;
  stream_rd_odd = 0;
  USEX(UART_DIV) = 0x006a; //460800bps
  USEX(INT_ENABLE) = 0x23; //enable DAC, SCI, RX
  USEX(SER_DREQ) = 1;
  USEX(INT_GLOB_ENA) = 1; //Needed if started from SCI

#if 0
  USEX(GPIO_DDR)   |= (1<<3); //GPIO3 to output and high
  USEX(GPIO_ODATA) |= (1<<3);
#endif

  USEX(TIMER_CONFIG) = 16-1; //2*USEX(&clockX)-1; //count 6MHz cycles
  /* Start a down-count timer to count the time between IMA ADPCM frames. */
  PERIP32(TIMER_T0L) = PERIP32(TIMER_T0CNTL) = 0x0000ffff;
  USEX(TIMER_ENABLE) = (1<<0);

  {
#define MIXER_OUT_SAMPLES 34
    struct Wav {
      s_int16 channels;
      u_int16 sampleRate;
      s_int16 ima;
      s_int16 bitsPerSample;
      s_int16 samplesPerBlock;
      u_int16 imaSync;
      u_int32 samplesLeft;
      u_int16 flags;
    };
    extern struct Wav wav;
    extern s_int16 mixerOut[3][2*MIXER_OUT_SAMPLES];
    int synced = 0;
    static s_int16 adpcmSt[2], adpcmVal[2];
    skipMixer = 1;
    adpcmSource = 0;    /* ADPCM from SDI */

#if ENCODE_CHANNELS != 1
    error('Only one-channel IMA supported!');
#else
    memsetY(audio_buffer, 0, 2048);
    SetHardware(ENCODE_CHANNELS, ENCODE_RATE);
    adpcmPhase = 0;
    while (1) {
      if (AudioBufFill() < 16) {
        /* Underflow -
           vs1103 doesn't have a DAC interrupt with underflow detection. */
        memset(mixerOut[0], 0, 2*16);
        WmaStereoCopy(mixerOut[0], 16);
      }
      if (StreamDiff() > 4/2) {
        void *old = stream_rd_pointer;
        AdpcmReadBytes(adpcmTmpX, 0, 4);
        /* Copy to real input buffer */
        adpcmTmpY[0] = adpcmTmpX[0];
        adpcmTmpY[1] = adpcmTmpX[1];
        
        if (0 == adpcmPhase) {
          if ((adpcmTmpX[1] & 255) != 0 ||
              adpcmTmpX[1] > (ISSTMAX<<8)) {
            stream_rd_pointer = old;
            //Wrong sync. Restore, skip, then try again.
            SdiGetC();
            synced = 0;
            continue;
          }
          if (++synced > 1) {
            /* Automatically determine rate:
               What's the time between two frames? */
            {
              u_int16 t = 0xffff - USEX(TIMER_T0CNTL);
              USEX(TIMER_T0CNTL) = 0xffff;
              t = (3072000 * 505) / t;
              USEX(SCI_AICTRL2) = t;
              if (t >= 46080U) {
        	t = 48000U;
              } else if (t >= 38000U) {
        	t = 44100U;
              } else if (t >= 28000U) {
        	t = 32000U;
              } else if (t >= 20160U) {
        	t = 24000U;
              } else if (t >= 14000U) {
        	t = 16000U;
              } else if (t >= 10500U) {
        	t = 12000U;
              } else {
        	t = 8000U;
              }
              if (t != hwSampleRate) {
        	SetHardware(1, t);
              }
            }
            AdjustFilter();
            ImaExpandStart(adpcmSt, adpcmVal, (__mem_y void *)adpcmTmpY, (void *)adpcmTmpX);
            mixerOut[0][0] = adpcmTmpX[0];
            mixerOut[0][1] = adpcmTmpX[0];
            WmaStereoCopy(mixerOut[0], 1);
          } else {
            USEX(TIMER_T0CNTL) = 0xffff;
          }
          adpcmPhase += 1;
        } else {
          if (synced > 1) {
            ImaExpandContinue(8, adpcmSt, adpcmVal, (__mem_y void *)adpcmTmpY, (void *)adpcmTmpX);
            {
              register s_int16 *d = mixerOut[0];
              register s_int16 *s = (void *)adpcmTmpX;
              register int i;
              for (i=0;i<8;i++) {
        	*d++ = *s;
        	*d++ = *s++;
              }
            }
            WmaStereoCopy(mixerOut[0], 8);
          }
          adpcmPhase += 8;
          if (adpcmPhase >= 505) {
            adpcmPhase = 0;
          }
          
        }
      }
    }
#endif
  }
}
